"""
    Evaluate a pipeline with metrics PSNR, SSIM, MSE 
    Over a preprocessed dataset
"""
import argparse
from pathlib import Path
from tqdm import tqdm 
import time 

from torch.utils.data import DataLoader
from skimage.metrics import (
    peak_signal_noise_ratio, 
    mean_squared_error,
    structural_similarity
)
import numpy as np

from pipeline import LDM_pipeline, DDRM_pipeline, Baseline_pipeline
from dataset import DegradDataset
from utils import data_transform, inverse_data_transform, np_to_pil

DEGRADATIONS = ['sr_bicubic_4', 'deblur_gauss_1', 'deblur_gauss_3', 'color', 'deno_0.1', 'deno_0.3', 'sr_bicubic_4_noisy_0.05']

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    
    # ----- Dataset ----- #
    parser.add_argument("--input", "-i", type=str, default="../data/preprocessed_imagenet-val-1k")
    parser.add_argument("--degrad", type=str, choices=DEGRADATIONS, default="sr_bicubic_4")
    parser.add_argument("--size", type=int, default=256)
    parser.add_argument("--length", type=int)
    # ----- Inference ----- #
    parser.add_argument("--pipeline", type=str, choices=["LDM", "DDRM", "DDRM_LDM", "Baseline"], default="LDM")
    parser.add_argument("--batch_size", type=int, default=16)
    parser.add_argument("--device", type=str, default="cuda")
    parser.add_argument("--output_dir", type=lambda x: Path(x).absolute(), default=None)
    parser.add_argument("--num_inference_steps", type=int, default=100)
    parser.add_argument("--get_intermediate_results", action='store_true')
    # ----- LDM pipelne ----- #
    parser.add_argument("--ldm_name", type=str, default="CompVis/ldm-super-resolution-4x-openimages")
    parser.add_argument("--ldm_eta", type=float, default=1)
    # ---- DDRM, DDRM_LDM pipeline ----- #
    parser.add_argument("--config_name", type=str, default="imagenet_256.yml")
    parser.add_argument("--ddrm_eta", type=float, default=0.85)
    parser.add_argument("--ddrm_etaB", type=float, default=1.0)
    args = parser.parse_args()

    if args.output_dir is None:
        if args.pipeline == 'LDM':
            pipeline_str = f"LDM-{args.num_inference_steps}"
        elif args.pipeline == "DDRM" or args.pipeline == "DDRM_LDM":
            pipeline_str = f"{args.pipeline}-{args.num_inference_steps}-{args.config_name.split('.')[0]}"
        else:
            pipeline_str = args.pipeline 
        dataset_str = args.input.split('_')[-1]
        if args.length is not None:
            dataset_str = f"{dataset_str}-{args.length}"
        name = f"{args.degrad}_{pipeline_str}_{dataset_str}"
        args.output_dir = Path("./result")/name

    pipeline_map = {"LDM": LDM_pipeline, "DDRM": DDRM_pipeline, "DDRM_LDM": DDRM_pipeline, "Baseline": Baseline_pipeline}
    basic_kwargs = {"return_intermediate_results": args.get_intermediate_results, "num_inference_steps": args.num_inference_steps, "device": args.device}
    ldm_kwargs = {**basic_kwargs, "ldm_name": args.ldm_name, "eta": args.ldm_eta}
    ddrm_kwargs = {**basic_kwargs, "config_name": args.config_name, "degrad": args.degrad, 
                    "size": args.size, "eta": args.ddrm_eta, "etaB": args.ddrm_etaB}
    baseline_kwargs = {"size": args.size}
    kwargs_map = {"LDM": ldm_kwargs, "DDRM": ddrm_kwargs, "DDRM_LDM": ddrm_kwargs, "Baseline": baseline_kwargs}

    pipeline = pipeline_map[args.pipeline](**kwargs_map[args.pipeline])

    # Parameters related to degradation and pipeline
    input_size = int(args.size / 4) if args.pipeline == "LDM" \
                                        or args.degrad.startswith("sr_bicubic_4") else args.size
    input_channel = 1 if args.pipeline == "DDRM" and args.degrad == "color" else 3

    dataset = DegradDataset(args.input, args.degrad, input_size=input_size, input_channel=input_channel, length=args.length)
    dataloader = DataLoader(dataset, batch_size=args.batch_size)

    psnrs, ssims, mses, times = [], [], [], []
    (args.output_dir).mkdir(exist_ok=True)
    idx = 0
    for orig_batch, degrad_batch in tqdm(dataloader):
        degrad_batch = data_transform(degrad_batch)
        t1 = time.time()
        output, intermediate_results = pipeline(degrad_batch)
        t2 = time.time()
        orig_batch_np = orig_batch.numpy().transpose(0, 2, 3, 1)
        
        psnrs.extend([peak_signal_noise_ratio(orig_batch_np[i], output[i]) for i in range(len(output))])
        ssims.extend([structural_similarity(orig_batch_np[i], output[i], data_range=1.0, channel_axis=-1) for i in range(len(output))])
        mses.extend([mean_squared_error(orig_batch_np[i], output[i]) for i in range(len(output))])
        times.append(t2 - t1)
        if args.get_intermediate_results:
            for i in range(output.shape[0]):
                for step in range(len(intermediate_results)):
                    tmp = np_to_pil(intermediate_results[step][i])
                    tmp.save(args.output_dir/f"{idx}-{step}.png")
                idx += 1
        else:
            for i in range(output.shape[0]):
                tmp = np_to_pil(output[i])
                tmp.save(args.output_dir/f"{idx}.png")
                idx += 1

    print("PSNR:", np.mean(psnrs))
    print("SSIM:", np.mean(ssims))
    print("MSE:", np.mean(mses))
    print("Total pipeline time:", np.sum(times))

