"""
A LDM module that can be used in DDRM pipeline
"""
from diffusers import LDMSuperResolutionPipeline # version 0.11.1
import torch.nn.functional as F
import torch
from diffusers.schedulers import DDIMScheduler

class CustomDDIMScheduler(DDIMScheduler):
    def reverse_step(self, prev_sample, sample, timestep, eta=0):
        """
        Recover noise from prev_sample (x_t-1) and sample (xt)
        Reverse function of DDIM scheduler step
        """
        # 1. get previous step value (=t-1)
        prev_timestep = timestep - self.config.num_train_timesteps // self.num_inference_steps

        # 2. compute alphas, betas
        alpha_prod_t = self.alphas_cumprod[timestep]
        alpha_prod_t_prev = self.alphas_cumprod[prev_timestep] if prev_timestep >= 0 else self.final_alpha_cumprod

        beta_prod_t = 1 - alpha_prod_t
        
        # 3. compute variance: "sigma_t(η)" -> see formula (16)
        # σ_t = sqrt((1 − α_t−1)/(1 − α_t)) * sqrt(1 − α_t/α_t−1)
        variance = self._get_variance(timestep, prev_timestep)
        std_dev_t = eta * variance ** (0.5)

        # 4. compute noise of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        noise = prev_sample - (alpha_prod_t_prev ** 0.5) / (alpha_prod_t ** 0.5) * sample / ((1 - alpha_prod_t_prev - std_dev_t ** 2) ** 0.5 - ((1 - alpha_prod_t) * alpha_prod_t_prev / alpha_prod_t) ** 0.5)
        
        return noise

class DDRM_LDM_model:
    def __init__(self, config, size, y_size, device="cuda"):
        self.pipeline = LDMSuperResolutionPipeline.from_pretrained(config.model_name)
        # self.pipeline.scheduler = CustomDDIMScheduler()
        self.pipeline = self.pipeline.to(device)
        self.size = size # target image size
        self.scale = config.scale
        self.y_size = y_size

    def __call__(self, x, image, t):
        image = image.view(-1, 3, self.y_size, self.y_size)
        image = F.interpolate(image, size=self.size//self.scale, mode='bicubic')
        latents = self.pipeline.vqvae.encoder(x) # (1, 3, 64, 64)
        latents_input = torch.cat([latents, image], dim=1) # (1, 6, 64, 64)
        latents_input = self.pipeline.scheduler.scale_model_input(latents_input, t)
        noise_pred = self.pipeline.unet(latents_input, t).sample # (1, 3, 64, 64)

        latents = self.pipeline.scheduler.step(noise_pred, int(t[0]), latents, eta=1).prev_sample
        pred_image = self.pipeline.vqvae.decode(latents).sample 
        et = x - pred_image
        # et = self.pipeline.scheduler.reverse_step(pred_image, x, int(t[0]), eta=0)
        return et

        

