# ----- Ref: https://github.com/bahjat-kawar/ddrm     ----- #
# ----- Paper: Denoising Diffusion Restoration Models ----- #
import os
import yaml
import argparse
import numpy as np
import torch
from models.utils import create_model, download
from svd_replacement import (
    SuperResolution,
    SRConv,
    Deblurring,
    Inpainting,
    Colorization,
    Denoising
)
from preprocess import bicubic_kernel
from models.denoising import efficient_generalized_steps
from utils import inverse_data_transform
from .DDRM_LDM_model import DDRM_LDM_model

def get_beta_schedule(beta_schedule, *, beta_start, beta_end, num_diffusion_timesteps):
    def sigmoid(x):
        return 1 / (np.exp(-x) + 1)

    if beta_schedule == "quad":
        betas = (
            np.linspace(
                beta_start ** 0.5,
                beta_end ** 0.5,
                num_diffusion_timesteps,
                dtype=np.float64,
            )
            ** 2
        )
    elif beta_schedule == "linear":
        betas = np.linspace(
            beta_start, beta_end, num_diffusion_timesteps, dtype=np.float64
        )
    elif beta_schedule == "const":
        betas = beta_end * np.ones(num_diffusion_timesteps, dtype=np.float64)
    elif beta_schedule == "jsd":  # 1/T, 1/(T-1), 1/(T-2), ..., 1
        betas = 1.0 / np.linspace(
            num_diffusion_timesteps, 1, num_diffusion_timesteps, dtype=np.float64
        )
    elif beta_schedule == "sigmoid":
        betas = np.linspace(-6, 6, num_diffusion_timesteps)
        betas = sigmoid(betas) * (beta_end - beta_start) + beta_start
    else:
        raise NotImplementedError(beta_schedule)
    assert betas.shape == (num_diffusion_timesteps,)
    return betas

def dict2namespace(config):
    namespace = argparse.Namespace()
    for key, value in config.items():
        if isinstance(value, dict):
            new_value = dict2namespace(value)
        else:
            new_value = value
        setattr(namespace, key, new_value)
    return namespace

def get_H_funcs(degrad, size=256, device="cuda"):
    if degrad == 'sr_bicubic_4':
        factor = 4
        k = np.zeros((factor * 4))
        for i in range(factor * 4):
            x = (1/factor)*(i- np.floor(factor*4/2) +0.5)
            k[i] = bicubic_kernel(x)
        k = k / np.sum(k)
        kernel = torch.from_numpy(k).float().to(device)
        H_funcs = SRConv(kernel / kernel.sum(), \
                            3, size, device, stride=factor)
    elif degrad.startswith('deblur_gauss_'):
        sigma = int(degrad.split('_')[-1])
        pdf = lambda x: torch.exp(torch.Tensor([-0.5 * (x/sigma)**2]))
        radius = np.ceil(5 * sigma).astype(int)
        kernel = torch.Tensor([pdf(r) for r in range(-radius, radius + 1)]).to(device)
        H_funcs = Deblurring(kernel / kernel.sum(), 3, 256, device)
    elif degrad == 'color':
        H_funcs = Colorization(size, device)
    elif degrad.startswith('deno_'):
        H_funcs = Denoising(3, size, device)
    elif degrad == 'sr_bicubic_4_noisy_0.05':
        factor = 4
        sigma_0 = 0.05
        k = np.zeros((factor * 4))
        for i in range(factor * 4):
            x = (1/factor)*(i- np.floor(factor*4/2) +0.5)
            k[i] = bicubic_kernel(x)
        k = k / np.sum(k)
        kernel = torch.from_numpy(k).float().to(device)
        H_funcs = SRConv(kernel / kernel.sum(), \
                            3, size, device, stride=factor)
    else:
        raise ValueError(f"Invalid degradation: {degrad}")
    return H_funcs 

def get_sigma_0(degrad):
    if degrad.startswith("deno_"): # deno_X
        sigma_0 = float(degrad.split("_")[-1])
    elif "noisy" in degrad: # 'sr_bicubic_4_noisy_0.05'
        sigma_0 = float(degrad.split("_")[-1])
    else:
        sigma_0 = 0.
    return sigma_0

class DDRM_pipeline:
    def __init__(self, 
                config_name="imagenet_256.yml",
                degrad="sr_bicubic_4",
                size=256,
                eta=0.85,
                etaB=1.0,
                num_inference_steps=100,
                return_intermediate_results=False,
                device="cuda"):
         # parse config file
        with open(os.path.join("configs", config_name), "r") as f:
            config = yaml.safe_load(f)
        config = dict2namespace(config)
        self.size = size
        self.eta = eta
        self.etaB = etaB
        self.num_inference_steps = num_inference_steps
        self.return_intermediate_results = return_intermediate_results
        self.device = device

        self.H_funcs = get_H_funcs(degrad)
        self.sigma_0 = get_sigma_0(degrad)

        betas = get_beta_schedule(
            beta_schedule=config.diffusion.beta_schedule,
            beta_start=config.diffusion.beta_start,
            beta_end=config.diffusion.beta_end,
            num_diffusion_timesteps=config.diffusion.num_diffusion_timesteps,
        )
        betas = self.betas = torch.from_numpy(betas).float().to(self.device)
        self.num_timesteps = betas.shape[0]

        alphas = 1.0 - betas
        alphas_cumprod = alphas.cumprod(dim=0)
        self.alphas_cumprod_prev = torch.cat(
            [torch.ones(1).to(device), alphas_cumprod[:-1]], dim=0
        )

        if config.model.type == 'openai':
            config_dict = vars(config.model)
            model = create_model(**config_dict)
            if config.model.use_fp16:
                model.convert_to_fp16()
            ckpt = os.path.join("checkpoints", "256x256_diffusion_uncond.pt")
            if not os.path.exists(ckpt):
                download('https://openaipublic.blob.core.windows.net/diffusion/jul-2021/256x256_diffusion_uncond.pt', ckpt)
                
            
            model.load_state_dict(torch.load(ckpt, map_location=self.device))
            model.to(self.device)
            model.eval()
            model = torch.nn.DataParallel(model)
        elif config.model.type == 'ldm':
            y_size = size // 4 if degrad.startswith("sr_bicubic_4") else size
            model = DDRM_LDM_model(config.model, size, y_size, device=self.device)
            model.pipeline.scheduler.set_timesteps(num_inference_steps, device=device)
        else:
            raise ValueError(f"Invalid model type: {config.model.type}")
        
        self.model = model
        self.model_type = config.model.type

    def __call__(self, batch):
        batch = batch.to(self.device)
        batch = batch.view(batch.shape[0], -1)
        ## Begin DDIM
        x = torch.randn(
            batch.shape[0],
            3,
            self.size,
            self.size,
            device=self.device,
        )
        skip = self.num_timesteps // self.num_inference_steps
        seq = range(0, self.num_timesteps, skip)

        with torch.no_grad():
            outputs, _ = efficient_generalized_steps(x, seq, self.model, self.betas, self.H_funcs, batch, self.sigma_0, \
                etaB=self.etaB, etaA=self.eta, etaC=self.eta, model_type=self.model_type)
        outputs = [inverse_data_transform(output) for output in outputs] # len: num_inference_steps + 1
        last_output = outputs[-1].cpu().permute(0, 2, 3, 1).numpy()
        if self.return_intermediate_results:
            return last_output, [output.cpu().permute(0, 2, 3, 1).numpy() for output in outputs[1:]] # discard the initial xs
        return last_output, None

if __name__ == '__main__':
    from dataset import DegradDataset
    from torch.utils.data import DataLoader

    config_name = "imagenet_256.yml"
    degrad = "color"
    size = 256
    pipeline_name = "DDRM"
    batch_size = 16
    
    pipeline = DDRM_pipeline(config_name, degrad)
    input_size = int(size / 4) if degrad.startswith("sr_bicubic_4") else size
    input_channel = 1 if pipeline_name == "DDRM" and degrad == "color" else 3
    
    dataset = DegradDataset("../data/preprocessed_imagenet-val-2k", degrad, input_size=input_size, input_channel=input_channel)
    dataloader = DataLoader(dataset, batch_size=batch_size)
    orig_batch, degrad_batch = next(iter(dataloader))
    pipeline(degrad_batch)