"""
    Preprocess ImageNet-1k dataset
    Convert images into target resolution
    By resizing at short edges and croping at long edges
    Downsample to generate low resolution images
    Store all files as lr_SIZE_ID.npy and hr_SIZE_ID.npy
"""
import argparse
from p_tqdm import p_map
from functools import partial
from tqdm import tqdm
from pathlib import Path
from PIL import Image
import PIL
import numpy as np
from dataset import ImageDataset
import matplotlib.pyplot as plt
from svd_replacement import (
    SuperResolution,
    SRConv,
    Deblurring,
    Inpainting,
    Colorization,
    Denoising
)
import torch
from utils import np_to_pil, data_transform, inverse_data_transform

DEGRADATIONS = ['sr_bicubic_4', 'deblur_gauss_1', 'deblur_gauss_3', 'color', 'deno_0.1', 'deno_0.3', 'sr_bicubic_4_noisy_0.05'] # or 'all'
device = "cuda"

def bicubic_kernel(x, a=-0.5):
    if abs(x) <= 1:
        return (a + 2)*abs(x)**3 - (a + 3)*abs(x)**2 + 1
    elif 1 < abs(x) and abs(x) < 2:
        return a*abs(x)**3 - 5*a*abs(x)**2 + 8*a*abs(x) - 4*a
    else:
        return 0

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    
    parser.add_argument('--img_dir', '-i', type=str, default="../data/imagenet-val-1k")
    parser.add_argument('--size', type=int, default=256)
    parser.add_argument('--degrad', type=str, nargs='+', default=['sr_bicubic_4'])
    parser.add_argument('--output', type=lambda x: Path(x).absolute(), required=True)
    
    args = parser.parse_args()

    if 'all' in args.degrad:
        args.degrad = DEGRADATIONS
    dataset = ImageDataset(args.img_dir, args.size)
    
    args.output.mkdir(exist_ok=True)
    (args.output/"original").mkdir(exist_ok=True)
    for degrad in args.degrad:
        (args.output/degrad).mkdir(exist_ok=True)
    
    for idx, data in tqdm(enumerate(dataset)):
        img_pil = np_to_pil(data.numpy().transpose(1, 2, 0))
        img_pil.save(args.output/"original"/f"orig_{idx}.png")
        data = data.to(device)
        data = data_transform(data)
        for degrad in args.degrad:
            if degrad == 'sr_bicubic_4':
                factor = 4
                k = np.zeros((factor * 4))
                for i in range(factor * 4):
                    x = (1/factor)*(i- np.floor(factor*4/2) +0.5)
                    k[i] = bicubic_kernel(x)
                k = k / np.sum(k)
                kernel = torch.from_numpy(k).float().to(device)
                H_funcs = SRConv(kernel / kernel.sum(), \
                                 3, args.size, device, stride=factor)
                y = inverse_data_transform(H_funcs.H(data.unsqueeze(0)))
                y = y.view(3, args.size//4, args.size//4).cpu().numpy().transpose(1, 2, 0)
                y_pil = np_to_pil(y)
                y_pil.save(args.output/degrad/f"{degrad}_{idx}.png")
                
            elif degrad.startswith('deblur_gauss_'):
                sigma = int(degrad.split('_')[-1])
                pdf = lambda x: torch.exp(torch.Tensor([-0.5 * (x/sigma)**2]))
                radius = np.ceil(5 * sigma).astype(int)
                kernel = torch.Tensor([pdf(r) for r in range(-radius, radius + 1)]).to(device)
                H_funcs = Deblurring(kernel / kernel.sum(), 3, 256, device)
                y = inverse_data_transform(H_funcs.H(data.unsqueeze(0)))
                y = y.view(3, args.size, args.size).cpu().numpy().transpose(1, 2, 0)
                y_pil = np_to_pil(y)
                y_pil.save(args.output/degrad/f"{degrad}_{idx}.png")
            
            elif degrad == 'color':
                H_funcs = Colorization(args.size, device)
                y = inverse_data_transform(H_funcs.H(data.unsqueeze(0)))
                y = y.view(y.shape[0], 1, args.size, args.size).repeat(1, 3, 1, 1).cpu().numpy().squeeze().transpose(1, 2, 0)
                y_pil = np_to_pil(y)
                y_pil.save(args.output/degrad/f"{degrad}_{idx}.png")

            elif degrad.startswith('deno_'):
                sigma_0 = float(degrad.split('_')[-1])
                y = data + sigma_0 * torch.randn_like(data)
                y = inverse_data_transform(y)
                y = y.cpu().numpy().squeeze().transpose(1, 2, 0)
                y_pil = np_to_pil(y)
                y_pil.save(args.output/degrad/f"{degrad}_{idx}.png")

            elif degrad == 'sr_bicubic_4_noisy_0.05':
                factor = 4
                sigma_0 = 0.05
                k = np.zeros((factor * 4))
                for i in range(factor * 4):
                    x = (1/factor)*(i- np.floor(factor*4/2) +0.5)
                    k[i] = bicubic_kernel(x)
                k = k / np.sum(k)
                kernel = torch.from_numpy(k).float().to(device)
                H_funcs = SRConv(kernel / kernel.sum(), \
                                 3, args.size, device, stride=factor)
                y = H_funcs.H(data.unsqueeze(0)).view(3, args.size//4, args.size//4)
                y = y + sigma_0 * torch.randn_like(y)
                y = inverse_data_transform(y)
                y = y.cpu().numpy().transpose(1, 2, 0)
                y_pil = np_to_pil(y)
                y_pil.save(args.output/degrad/f"{degrad}_{idx}.png")
                
            else:
                raise ValueError(f"Invalid degradation: {degrad}")

        
    